/*  umidi, MIDI to URBI bridge
 *   Copyright (C) 2009  Matthieu Nottale
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 3 as
 *   published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "midiin.hh"

#ifdef WIN32
std::vector<std::string> MidiIn::listDevices()
{
  std::vector<std::string> res;
  for (unsigned int i=0; i< midiInGetNumDevs(); i++)
  {
    MIDIINCAPS caps;
    midiInGetDevCaps(i, &caps, sizeof(MIDIINCAPS));
    res.push_back(caps.szPname);
  }
  return res;
}
void CALLBACK MidiInProc(HMIDIIN, UINT wMsg, DWORD_PTR inst,
  DWORD_PTR dwParam1, DWORD_PTR dwParam2
)
{
  MidiIn* dev = (MidiIn*)inst;
  dev->rawData(wMsg, dwParam1, dwParam2);
}
bool MidiIn::open(int devId)
{
  return midiInOpen(&handle, devId,
    reinterpret_cast<DWORD>(MidiInProc),
    reinterpret_cast<DWORD>(this),
    CALLBACK_FUNCTION) == MMSYSERR_NOERROR &&
  midiInStart(handle) == MMSYSERR_NOERROR;
}

void MidiIn::rawData(UINT msg, DWORD p1, DWORD)
{
  if (debug)
    std::cerr << "message " << msg << " " << p1 << std::endl;
  if (msg == MIM_DATA)
  {
    // http://msdn.microsoft.com/en-us/library/ms712073(VS.85).aspx
    unsigned char chan = p1 & 0x0F;
    unsigned char msg = (p1 & 0xF0) >> 4;
    unsigned char v1 = (p1 & 0xFF00) >> 8;
    unsigned char v2 = (p1 & 0xFF0000) >> 16;
    midiData(msg, chan, v1, v2);
  }
  if (msg == MIM_LONGDATA)
    std::cerr <<"No idea what to do..." << std::endl;
}
#else
#include <alsa/asoundlib.h>
#include <libport/thread.hh>
std::vector<std::string> MidiIn::listDevices()
{
  std::vector<std::string> res;
  return res;
}


bool MidiIn::open(int devId)
{
   int err;
   err = snd_seq_open(&handle, "default", SND_SEQ_OPEN_INPUT, 0);
   if (err < 0)
     return false;
   snd_seq_set_client_name(handle,"midiin");
   int port = snd_seq_create_simple_port(handle, "input",
     SND_SEQ_PORT_TYPE_MIDI_GENERIC |
     SND_SEQ_PORT_TYPE_SOFTWARE |
     SND_SEQ_PORT_CAP_WRITE |
     SND_SEQ_PORT_CAP_SUBS_WRITE,
     SND_SEQ_PORT_TYPE_SYNTH
     );
   libport::startThread(this);
   return true;
}

void MidiIn::operator()()
{
  while(handle)
  {
     snd_seq_event_t* event=0;
     // Blocking call.
     snd_seq_event_input(handle, &event);
     if (!event)
       continue;
     unsigned char msg,chan, p1, p2;
     switch(event->type)
     {
     case SND_SEQ_EVENT_NOTEON:
     case SND_SEQ_EVENT_NOTEOFF:
       msg = (event->type == SND_SEQ_EVENT_NOTEON)? NOTE_ON:NOTE_OFF;
       chan = event->data.note.channel;
       p1=event->data.note.note;
       p2=event->data.note.velocity;
       break;
     case SND_SEQ_EVENT_PITCHBEND:
       msg = PITCH_WHEEL;
       chan = event->data.control.channel;
       p1 = 0;
       p2 = event->data.control.value;
       break;
     case SND_SEQ_EVENT_CONTROLLER:
       msg = PARAMETER;
       chan = event->data.control.channel;
       p1 = event->data.control.param;
       p2 = event->data.control.value;
       break;
     case SND_SEQ_EVENT_CLOCK:
       msg = REALTIME;
       chan = TICK;
       break;
     case SND_SEQ_EVENT_START:
       msg = REALTIME;
       chan = START;
       break;
     case SND_SEQ_EVENT_STOP:
       msg = REALTIME;
       chan = STOP;
       break;
     case SND_SEQ_EVENT_CONTINUE:
       msg = REALTIME;
       chan = CONTINUE;
       break;
     case SND_SEQ_EVENT_RESET:
       msg = REALTIME;
       chan = RESET;
       break;
     case SND_SEQ_EVENT_SENSING:
       msg = REALTIME;
       chan = SENSE;
       break;
     default:
       std::cerr <<"unknown msg " << (int)event->type << std::endl;
       continue;
     }
     midiData(msg, chan, p1, p2);
  }
}

#endif
void MidiIn::dump(std::ostream& o, unsigned char msg, unsigned char channel,
    unsigned char v1, unsigned char v2)
{
  static const char* mt[8] = {"Note off", "Note on", "key pressure",
  "parameter", "program", "chan pressure", "pitch wheel", "realtime"};
  static const char* rt[8] = {"clocktick", "undef", "start", "continue",
  "stop", "undef", "sensing", "reset"};
  o << (int)msg << ' ' << mt[msg & 7] << ' ';
  if (msg == REALTIME)
    o << rt[channel & 7];
  else
    o << (int)channel;
  o << ' ' << (int)v1 << ' ' << (int)v2 << std::endl;
}
