/*  umidi, MIDI to URBI bridge
 *   Copyright (C) 2009  Matthieu Nottale
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 3 as
 *   published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/lexical_cast.hpp>
#include "umidi.hh"
#include "factory.hh"
inline bool isUp(unsigned char cmd, unsigned char vel)
{
  return cmd == MidiIn::NOTE_OFF || (cmd == MidiIn::NOTE_ON && vel == 0);
}

#define FACTORY(n)  \
Midi::Operation* n##create() {return new n;}  \
Factory<Midi::Operation> n##Factory(#n, &n##create)
namespace midiop
{
  class FillParams{
  public:
    FillParams(const Midi::Params& p)
    :p(p)
    , ok(true)
    , pos(0)
    , optional_(false)
    {
    }
    struct Optional{};
    static Optional optional;
    operator bool () const { return ok && pos==p.size();}
    const Midi::Params& p;
    bool ok;
    int pos;
    template<class T> FillParams& operator >>(T& t)
    {
      if (pos >= p.size())
      {
	if (optional_)
	  return *this;
	ok = false;
	return *this;
      }
      t = boost::lexical_cast<T>(p[pos]);
      pos++;
      return *this;
    }
    FillParams& operator >>(Optional&)
    {
      optional_ = true;
      return *this;
    }
    // Remaining parameters are optional.
    bool optional_;
  };
  FillParams::Optional FillParams::optional;
  class DirectMap: public Midi::Operation
  {
  public:
    enum What {
      PITCH,
      VELOCITY
    };
    virtual bool setup(const Midi::Params& p)
    {
      resetVal_ = -1;
      appendPitch_ = false;
      return FillParams(p) >> target_ >> what_ >> factor_ >> offset_
      >> FillParams::optional >> resetVal_ >> appendPitch_;
    }
    virtual void midiData(Midi::Target t, unsigned char msg, unsigned char,
    unsigned char v1, unsigned char v2)
    {
      double tval;
      if (isUp(msg, v2))
	if (resetVal_ == -1)
	  return;
	else
	  tval = resetVal_;
      else
	tval = (double)((What)what_ == PITCH ? v1:v2)*factor_ + offset_;
      std::string tgt = target_;
      if (appendPitch_)
	tgt += boost::lexical_cast<std::string>((int)v1);
      midi->writeVar(t, tgt, tval);
    }
  private:
    std::string target_;
    int what_;
    double factor_;
    double offset_;
    double resetVal_;
    bool appendPitch_;
  };
  FACTORY(DirectMap);

  class ParameterSet: public Midi::Operation
  {
  public:
    virtual bool setup(const Midi::Params& p)
    {
      return FillParams(p) >> bitMask_;
    }
    virtual void midiData(Midi::Target t, unsigned char msg, unsigned char,
    unsigned char v1, unsigned char v2)
    {
      bool up = isUp(msg, v2) || (msg==MidiIn::PARAMETER && v2);
      if (up)
	midi->setParameter(midi->getParameter() | bitMask_);
      else
	midi->setParameter(midi->getParameter() &~bitMask_);
    }
    private:
    unsigned long bitMask_;
  };
  FACTORY(ParameterSet);

  class EmitMap: public Midi::Operation
  {
  public:
    virtual bool setup(const Midi::Params& p)
    {
      if (p.size() != 1)
	return false;
      event_ = p.front();
      return true;
    }
    virtual void midiData(Midi::Target t, unsigned char msg,
      unsigned char channel, unsigned char v1, unsigned char v2)
    {
      std::stringstream ss;
      ss <<"emit " << event_ <<'(' << (int)msg << ',' << (int)channel
      << ',' << (int)v1 << ',' << (int)v2 <<"),";
      midi->sendCommand(t, ss.str());
    }
    std::string event_;
  };
  FACTORY(EmitMap);

  // Freeze tag on up, unfreeze while note down, will fail if multiples notes
  class TagMap: public Midi::Operation
  {
  public:
    virtual bool setup(const Midi::Params& p)
    {
      if (p.size() != 1 && p.size() !=2)
	return false;
      tag_ = p.front();
      appendPitch_ = (p.size() == 2 && (p[1]=="1" || p[1] == "true"));
      return true;
    }
    virtual void midiData(Midi::Target t, unsigned char msg, unsigned char,
    unsigned char v1, unsigned char v2)
    {
      if (midi->debug > 10)
	std::cerr <<"TagMap midiData" << std::endl;
      std::string cmd;
      if (!isUp(msg, v2))
	cmd = "un";
      cmd += "freeze " + tag_;
      if (appendPitch_)
	cmd += boost::lexical_cast<std::string>((int)v1);
      cmd +=",";
      midi->sendCommand(t, cmd);
    }
    std::string tag_;
    bool appendPitch_;
  };
  FACTORY(TagMap);
  // Switch tag state at each match
  class TagSwitch: public Midi::Operation
  {
  public:
    virtual bool setup(const Midi::Params& p)
    {
      return FillParams(p) >> tag_ >> state_;
    }
    virtual void midiData(Midi::Target t, unsigned char msg, unsigned char,
    unsigned char, unsigned char v2)
    {
      if (isUp(msg, v2))
	return;
      state_ = !state_;
      std::string cmd;
      if (state_)
	cmd = "un";
      cmd += "freeze " + tag_ + ",";
      midi->sendCommand(t, cmd);
    }
    std::string tag_;
    bool state_;
  };
  FACTORY(TagSwitch);
}
