/*  umidi, MIDI to URBI bridge
 *   Copyright (C) 2009  Matthieu Nottale
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 3 as
 *   published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sstream>
#include <stdexcept>
#include <boost/lexical_cast.hpp>

#include <libport/foreach.hh>
#include <libport/assert.hh>
#ifndef WIN32
#include <netinet/in.h>
#include <netinet/tcp.h>
typedef int ioctl_type;
#else
typedef char ioctl_type;
#endif
/* Put that axe back where you found it. I realy need to set TCP_NODELAY on the
 * UClient socket as we care a lot about latency.
 */
#define protected public
#include <urbi/uclient.hh>
#undef protected
#include "midiin.hh"
#include "umidi.hh"

Midi::Midi()
:debug(0)
,rtTicks(0)
,running(false)
,transmitTick(1)
,parameter(0)
{
  targets.resize(32,0);
}

int Midi::init(int dev)
{
  if (!open(dev))
    return -1;
  return 0;
}

int Midi::close()
{
  MidiIn::close();
  return 0;
}
int Midi::unregister(ChanMask mask, long ptr)
{
  libport::BlockLock bl(listLock);
  bool removed = false;
  void* addr = (void*)ptr;
  for (int c =0; c<32; c++)
    if (mask & (1<<c))
    {
      int chan = 1 << c;
      for (unsigned int p=0; p<actions[chan].size(); p++)
	if (actions[chan][p].operation == addr)
	{
	  if (p != actions[chan].size() - 1)
	    std::swap(actions[chan][actions[chan].size()-1], actions[chan][p]);
	  actions[chan].pop_back();
	  removed = true;
	  break;
	}
    }
    return removed?1:0;
}

int Midi::sendCommand(Target t, const std::string& cmd)
{
  if (debug > 9)
    std::cerr <<"Sending '" << cmd <<"' to " << t << std::endl;
  for (int r=0; r<32; r++)
    if ((t & (1 << r)) && targets[r])
      targets[r]->effectiveSend(cmd.c_str(), cmd.length());
  return 0;
}

long Midi::registerAction(ChanMask mask, Filter f, const std::string& name,
    const Params& ps, Target target)
{
  Operation* o = Factory<Operation>::create(name);
  if (!o)
  {
    if (debug > 1)
      std::cerr <<"Factory creation failed for " << name << std::endl;
    return 0;
  }
  o->init(this);
  if (!o->setup(ps))
  {
    if (debug > 2)
      std::cerr <<"Setup failed for " << name << std::endl;
    delete o;
    return 0;
  }
  if (debug > 3)
    std::cerr << "Installing a nnew " << name <<" on " << mask
  <<" with filter " << (int)f.pitch <<'/' << (int)f.pitchMask << std::endl;
  libport::BlockLock bl(listLock);
  for (int m=0; m<32;m++) if (mask & (1 << m))
  {
    if (debug > 10)
      std::cerr <<" on channel " << m << std::endl;
    actions[m].push_back(Action(f, o, target));
  }
  return (long)o;
}

/*
UList Midi::listDevices()
{
  std::vector<std::string> devs = MidiIn::listDevices();
  UList res;
  foreach(const std::string& s, devs)
  res.array.push_back(new UValue(s));
  return res;
}*/

void Midi::midiData(unsigned char msg, unsigned char channel,
    unsigned char v1, unsigned char v2)
{
  if ((debug>10 && msg!=REALTIME) || debug>10000)
    dump(std::cerr, msg, channel, v1, v2);
  if (msg == REALTIME)
    realTime(channel);
  if (msg != NOTE_OFF && msg != NOTE_ON && msg != PARAMETER)
    return;
  if (debug>5 && debug<10)
    dump(std::cerr, msg, channel, v1, v2);
  if (msg == PARAMETER)
    channel += 16;
  libport::BlockLock bl(listLock);
  if (debug > 10)
    std::cerr << actions[channel].size() <<" actions on channel "
    << (int)channel << std::endl;
  foreach(Action& a, actions[channel])
  {
    if (debug > 50)
      std::cerr << "trying " <<(int)v1 <<" on filter " << (int)a.filter.pitch
    << '/' <<(int)a.filter.pitchMask << '=' << a.filter.parameter
    <<'/' << a.filter.parameterMask << std::endl;
    if (a.filter.match(v1, parameter))
    {
      if (debug > 20)
	std::cerr << (long)a.operation << " has a matching filter, execing.."
      << std::endl;
      a.operation->midiData(a.target, msg, channel, v1, v2);
    }
  }
}

void Midi::realTime(unsigned char rmsg)
{
  if (debug > 1000)
    std::cerr <<"Realtime " << (int)rmsg <<" " << running << std::endl;
  switch(rmsg | 0x8)
  {
  case TICK:
    /* Some devices dont care much about running mode...
    if (!running)
      std::cerr << "Warning, unexpected tick in stopped state" << std::endl;
    */
    if (transmitTick && (! (rtTicks % (24 / transmitTick))))
      sendCommand(0xFFFFFFFF, "emit midi.tick( "
        + boost::lexical_cast<std::string>(rtTicks) + "),");
    if (!rtTicks && mStartTime)
    {
      libport::utime_t period = libport::utime() - mStartTime;
      writeVar(0xFFFFFFFF,  "midi.bpm", 60000000 / period);
      writeVar(0xFFFFFFFF,  "midi.period",  period / 1000);
    }
    if (!rtTicks)
      mStartTime = libport::utime();
    rtTicks++;
    rtTicks %= 24;
    break;
  case START:
    if (running)
    {
      std::cerr <<"Extra start message" << std::endl;
      break;
    }
    mStartTime = 0;
    running = true;
    sendCommand(0xFFFFFFFF, "emit midi.start(),");
    break;
  case STOP:
    if (!running)
      break;
    running = false;
    rtTicks = 0;
    sendCommand(0xFFFFFFFF, "emit midi.stop(),");
    break;
  }
}
static urbi::UCallbackAction dump(const urbi::UMessage& m)
{
  std::cerr <<"Urbi error: " << m.rawMessage << std::endl;
  return urbi::URBI_CONTINUE;
}
int Midi::registerServer(int id, const std::string& host, int port)
{
  if (id<0 || id >=32)
    return -1;
  urbi::USyncClient* cl = new urbi::USyncClient(host, port);
  cl->setCallback(urbi::callback(*this, &Midi::urbiMessage), "midictl");
  cl->setErrorCallback(urbi::callback(&::dump));
  cl->send("%s", "var midi.connectionID = connectionID;"
    "function midi.setDebug(p1) {"
    " if (connectionID == midi.connectionID)"
    "    midictl << [\"debug\", p1]"
    " else {"
    "   p.p1=x; +connection(\"all\"):"
    "   {midictl << [\"debug\",p.p1]}}};"
    "function midi.setParameter(p1) {"
    " if (connectionID == midi.connectionID) midictl << [\"setparameter\",p1]"
      " else {p.p1=p1; +connection(\"all\"): "
    "{midictl << [\"setparameter\",p.p1]} | noop | noop}}; "
    "function midi.registerAction(p1, p2, p3, p4, p5) {"
    " if (connectionID == midi.connectionID)"
    "   midictl << [\"register\", p1, p2, p3, p4,p5]"
    "  else {p.p1=p1| p.p2 = p2|p.p3 = p3| p.p4=p4| p.p5=p5|"
    "+connection(\"all\"): {midictl << [\"register\", p.p1, p.p2, p.p3, p.p4,"
    "p.p5]} | noop | noop;}};"
    "var midi.connected = 1;");
  cl->sendFile("umidi.u");
  //Minimise latency
  ioctl_type c = 1;
  if (setsockopt(cl->sd, IPPROTO_TCP, TCP_NODELAY, &c, sizeof(ioctl_type)))
    libport::perror("setsockopt");
  targets[id] = cl;
  return 1;
}

class ListExtractor {
public:
  ListExtractor(const urbi::UList& l)
  :l(l)
  ,pos(0)
  ,ok(true){}
  operator bool () { return ok;}
#define VALOP(t) ListExtractor& operator >>(t& v)                         \
  { if (!ok) return *this;                                                  \
    if (l.array.size()<=pos ||  l.array[pos]->type != urbi::DATA_DOUBLE)    \
      ok=false; else                                                        \
         v = l.array[pos]->val; pos++; return *this;}

  VALOP(double)
  VALOP(int)
  VALOP(unsigned int)
  VALOP(unsigned long)
  VALOP(char)
  VALOP(unsigned char)
  #undef VALOP
  #define OP(ctype, enumtype, valfield)  \
  ListExtractor& operator >> (ctype & v) { \
    if(!ok) return *this; \
    if (l.array.size()<=pos || l.array[pos]->type != enumtype) ok = false; \
    else v = *l.array[pos]->valfield; \
  pos++; return *this;}
  OP(std::string, urbi::DATA_STRING, stringValue)
  OP(urbi::UList, urbi::DATA_LIST, list)
  #undef OP
private:
  const urbi::UList& l;
  int pos;
  bool ok;
};

urbi::UCallbackAction Midi::urbiMessage(const urbi::UMessage& m)
{
  if (m.type == urbi::MESSAGE_DATA && m.value->type == urbi::DATA_LIST
    && m.value->list->array.size() >= 1
    && m.value->list->array[0]->type == urbi::DATA_STRING)
  {
    urbi::UList& l = *m.value->list;
    std::string key = *l.array[0]->stringValue;
    if (key == "debug")
    {
      if (l.size() == 2 && l.array[1]->type == urbi::DATA_DOUBLE)
	debug = m.value->list->array[1]->val;
      std::cerr <<"Set debug to " << debug << std::endl;
    }
    if (key == "setparameter")
    {
      unsigned long p;
      if (ListExtractor(l) >> p)
      {
	setParameter(p);
      }
    }
    if (key == "register")
    {
      unsigned int cmask, target;
      std::string name;
      urbi::UList params;
      urbi::UList lfilter;
      if (ListExtractor(l) >> key >> cmask  >> lfilter >> name >> params
	>> target)
      {
	Params p;
	foreach(urbi::UValue* v, params.array)
	{
	  if (v->type != urbi::DATA_STRING && v->type != urbi::DATA_DOUBLE)
	  {
	    std::cerr <<"params extraction failure" << std::endl;
	    break;
	  }
	  if (v->type == urbi::DATA_STRING)
	    p.push_back(*v->stringValue);
	  else
	    p.push_back(boost::lexical_cast<std::string>(v->val));
	}
	Filter f;
	//f2 is between 0 and 255
	unsigned long f1=0, f2=300, f3=0, f4=0;
	ListExtractor(lfilter) >> f1 >> f2 >> f3 >> f4;
	if (f2 == 300)
	  f2 = (f1==0?0:255);
	f.pitch = f1; f.pitchMask = f2; f.parameter = f3; f.parameterMask=f4;
	long a = registerAction(cmask, f, name, p, target);
	std::cerr << "Registering action: " << a << std::endl;
      }
      else
	std::cerr <<" parameter extraction failed" << std::endl;
    }
  }
  return urbi::URBI_CONTINUE;
}
static std::istream& operator >> (std::istream& is, Midi::Filter& f)
{
  // Careful, f.pitch is a char.
  int pitch=0, pitchMask=0;
  unsigned long parameter, parameterMask;
  is >> pitch;
  int c = is.get();
  if (c == '/')
  {
    is >> pitchMask;
    c = is.get();
    if (c == '=')
    {
      is >> parameter;
      c = is.get();
      if (c == '/')
	is >> parameterMask;
      else
	is.unget();
    }
    else
      is.unget();
  }
  else
  {
    is.unget();
    pitchMask = (pitch==0?0:0xFF);
  }
  f.pitch = pitch; f.pitchMask = pitchMask;
  f.parameter = parameter; f.parameterMask = parameterMask;
  return is;
}

bool Midi::parseConf(std::istream& is)
{
  if (!is.good())
    return false;
  char line[1024];
  while(true)
  {
    line[0] = 0;
    is.getline(line, 1024);
    if (!is.good())
      return true;
    if (line[0] == '#')
      continue;
    std::stringstream ss(line);
    std::string cmd;
    ss >> cmd;
    if (cmd == "robot")
    {
      std::string host;
      int port = 54000;
      int id = -1;
      ss >> id >> host >> port;
      if (id <0 || id >= 32)
	throw std::runtime_error("malformed line: " + (std::string)line);
      std::cerr <<"Registering robot " << host <<" " << port << std::endl;
      if (!registerServer(id, host, port))
	throw std::runtime_error("Failed to regiter robot " + host);
    }
    else if (cmd == "action")
    {
      unsigned int mask; Filter f; Target t; std::string name;
      ss >> std::hex >> mask;
      if (!ss.good())
	throw std::runtime_error("malformed mask: " + (std::string)line);
      ss >> std::dec >> f;
      if (!ss.good())
	throw std::runtime_error("malformed filter: " + (std::string)line);
      ss >> std::hex >> t >> name;
      if (!ss.good())
	throw std::runtime_error("malformed stuffs: " + (std::string)line);
      Params ps;
      while (true)
      {
	std::string p;
	ss >> p;
	if (p.empty())
	  break;
	ps.push_back(p);
      }
      if (!registerAction(mask, f, name, ps, t))
	throw std::runtime_error("Failed to register action from "
	  + (std::string)line);
    }
  }
  return true;
}

void Midi::setParameter(unsigned long p)
{
  parameter = p;
  std::cerr <<"Setting parameter to " << p << std::endl;
  writeVar(0xFFFFFFFF, "midi.parameter", p);
}


INSTANCIATE_FACTORY(Midi::Operation);
