/*  umidi, MIDI to URBI bridge
 *   Copyright (C) 2009  Matthieu Nottale
 *
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License version 3 as
 *   published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef UMIDI_HH
#define UMIDI_HH

#include <vector>
#include <string>
#include <libport/utime.hh>
#include <urbi/usyncclient.hh>
#include "factory.hh"
#include "midiin.hh"
class Midi: public MidiIn
{
public:
  struct Filter
  {
    Filter(unsigned char p, unsigned char pm)
    :pitch(p), pitchMask(pm) {}
    Filter() : pitch(0), pitchMask(0), parameter(0), parameterMask(0) {}
    unsigned char pitch;
    unsigned char pitchMask;
    unsigned long parameter;
    unsigned long parameterMask;
    inline bool match(unsigned char pitch, unsigned long parameter)
    {
      return  (pitch&pitchMask) == this->pitch
      && (parameter&parameterMask) == this->parameter;
    }
  };
  Midi();
  typedef std::vector<std::string> Params;
  // bit 0-15: note channels. bit 16-31: potar channels
  typedef unsigned long ChanMask;
  // bitmask on the robot list
  typedef unsigned long Target;
  Midi(const std::string&n);
  int init(int dev);
  int close();
  /// Unregister \b ptr from channels \b m.
  int unregister(ChanMask m, long ptr);
  /// Change the target of \b ptr on \b mask to \b t.
  //later int reTarget(ChanMask mask, long ptr, Target t);
  long registerAction(ChanMask mask, Filter f, const std::string& name,
    const Params& ps, Target target);

  int debug;
/* Interface for Operations: */
  /// Send command \b act to \b t
  int sendCommand(Target t, const std::string& act);
  /// Wrapper, sending the command varName = val
  template<typename T>
  int writeVar(Target t, const std::string& varName, const T& tval);

  int registerServer(int id, const std::string& host, int port = 54000);
  bool parseConf(std::istream& is);
  class Operation
  {
  public:
    virtual bool setup(const Params& p) {return true;}
    virtual void midiData(Target target, unsigned char msg,
      unsigned char channel, unsigned char v1, unsigned char v2) = 0;
    Operation* init(Midi* m) { midi = m; return this;}
  protected:
    Midi* midi;
  };
  //urbi control callback
  urbi::UCallbackAction urbiMessage(const urbi::UMessage&);
  virtual void midiData(unsigned char msg, unsigned char channel,
    unsigned char v1, unsigned char v2);
  /// 0:disable n: send n ticks per mesure (1 2 3 6 8 12 24)
  unsigned int transmitTick;
  /// Globaly available parameter state variable.
  unsigned long getParameter() {return parameter;}
  void setParameter(unsigned long p);
private:
  unsigned long parameter;
  void realTime(unsigned char rmsg);
  struct Action
  {
    Action(Filter filter, Operation* operation, Target target)
    :filter(filter), operation(operation),target(target) {}
    Filter filter;
    Operation* operation;
    Target target;
  };

  /// actions, sorted per channel
  std::vector<Action> actions[32];
  /// action list lock
  libport::Lockable listLock;
  /// List of target servers
  std::vector<urbi::USyncClient*> targets;
  /// number of received ticks
  unsigned int rtTicks;
  /// Status
  bool running;
  /// Last measure starts
  libport::utime_t mStartTime;
};
template<typename T>
int Midi::writeVar(Target t, const std::string& varName, const T& tval)
{
  std::stringstream ss;
  ss << varName << '=' << tval <<',';
  return sendCommand(t, ss.str());
}
#endif
